#include <stdio.h>
#include <stdarg.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>

extern "C" void *__libc_malloc(unsigned int size);
extern "C" void *__libc_calloc(unsigned int n, unsigned int size);
extern "C" void *__libc_free(void *pt);
extern "C" void *__libc_realloc(void *pt, unsigned int size);
extern "C" pid_t __libc_fork();

void show_trace_stub();
void show_trace();

static int fout=-1;
static bool init = false;
static const char *prefix;
//static int childnum=0;

static void show_init(const char *insert)
{
	if (fout == -1){
		init = true;
		prefix = getenv ("MAPMALLOCPREFIX");
		if (prefix == NULL) prefix = "/tmp/mapmalloc";
		char path[PATH_MAX];
		snprintf (path,sizeof(path)-1,"%s%s-%d.log",prefix,insert,getpid());
		fout = open (path,O_WRONLY|O_CREAT,0644);
		init = false;
	}
}

static void show_init ()
{
	show_init ("");
}

static char line[2000];
static char *pt=line;

static void show_printf (const char *ctl, ...)
{
	va_list list;
	va_start (list,ctl);
	pt += vsnprintf	(pt,200,ctl,list);
	va_end (list);
}

static void show_endline()
{
	*pt++ = '\n';
	write (fout,line,(int)(pt-line));
	pt = line;
}

void show_trace0()
{
	unsigned long data=10;
	unsigned long *stack = &data;
	int i;
	for (i=0; i<1000; i++){
		// printf ("stack=%08x %08x %08x %08x\n",stack,*stack,show_trace,main);
		if (*stack >= (unsigned long)show_trace
			&& *stack < (unsigned long)show_trace_stub) break;
		stack++;
	}
	if (i == 1000){
		show_printf ("Il ne trouve pas le stack");
		show_endline();
	}else{
		//printf ("trouve %08x %08x %08x %08x\n"
		//	,stack[-1],stack[0],stack[1],stack[2]);
		stack--;
		stack = (unsigned long*)stack[0];
		stack = (unsigned long*)stack[0];
		while (stack != NULL && stack >= (void*)0x1000){
			// printf ("bt: %08x %08x\n",stack[0],stack[1]);
			show_printf (" %08lx",stack[1]);
			stack = (unsigned long*)stack[0];
		}
	}
}

void show_trace()
{
	show_trace0();
}

void show_trace_stub(){}

extern "C" void *malloc (unsigned int size)
{
	void *ret = NULL;
	if (size != 0){
		ret = __libc_malloc(size);
		if (!init){
			show_init();
			show_printf ("malloc %08lx %d: ",(long)ret,size);
			show_trace();
			show_endline ();
		}
	}
	// fprintf (stderr,"bt %p %p %p\n",show_trace0,show_trace,malloc);
	return ret;
}

extern "C" void free(void *p)
{
	if (p != NULL){
		show_printf ("free %08lx:",(long)p);
		show_trace();
		show_endline();
		__libc_free (p);
	}
}

extern "C" void *realloc (void *pt, unsigned int size)
{
	void *ret = __libc_realloc (pt,size);
	show_printf ("realloc %08lx %08lx %u:",pt,ret,size);
	show_trace();
	show_endline();
	return ret;
}
	

extern "C" void *calloc (unsigned int n, unsigned int elm)
{
	void *ret = NULL;
	if (n != 0 && elm != 0){
		ret = __libc_calloc(n,elm);
		if (!init){
			show_init();
			show_printf ("calloc %08lx %d: ",(long)ret,n*elm);
			show_trace();
			show_endline ();
		}
	}
	return ret;
}

extern "C" pid_t fork()
{
	pid_t old = getpid();
	pid_t ret = __libc_fork();
	if (ret == 0){
		close (fout);
		fout = -1;
		char tmp[20];
		snprintf (tmp,sizeof(tmp)-1,"-%d-fork",old);
		show_init(tmp);
	}
	return ret;
}

